/*
 * Copyright 1996 by David Mosberger, Tucson, Arizona.
 *
 * This file is in the public domain.  Permission to use, copy,
 * modify, and distribute this software and its documentation for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice appear in all copies and that both that copyright
 * notice and this permission notice appear in supporting
 * documentation.
 *
 * Reads an ECOFF object file, schedules the code for an EV4 Alpha
 * chip (21064, 21066 and similar) and then prints the resulting issue
 * order and execution time.
 *
 * OPTIONS:
 *	-r To reorder (schedule) the code
 *	-u To toggle the alignment of the first instruction.  Normally,
 *	the first instruction is assumed to be 8 byte aligned, with this
 *	toggle, alignment can be switched to 4 bytes only
 *
 * CAVEATS:
 *	The code schedules BASIC BLOCKS only!  That means that if
 *	you have any kind of branch into or out of a block of code, the
 *	scheduler MAY NOT DO THE RIGHT THING.  If that happens, the timing
 *	information is wrong, in the best case.  In the worst case,
 *	rescheduling may result in bad code (e.g., it may incorrectly
 *	move a load past a branch target).
 *
 *	Also, the maximum size of a basic-block is hardcoded via macro
 *	MAX_NODES in file instr-sched.c.  The current value is 512 which
 *	should be plenty for most basic-blocks.  But there is no reason
 *	this constant can't be increased (memory requirements grow with the 
 *	square of that constant, so it may end up being a real memory hog
 *	when MAX_NODES is large).
 *
 */
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <aouthdr.h>
#include <filehdr.h>
#include <scnhdr.h>
#include <errno.h>


int
main (int argc, char *argv[])
{
    FILE *infp;
    long next_section;
    int i, j;
    int offset = 0;
    int reorder = 0;
    int *insts;
    struct filehdr fhdr;	/* file header */
    struct aouthdr ohdr;	/* optional header */
    struct scnhdr shdr;		/* section header */

    if (argc < 2) {
	fprintf(stderr, "usage: %s [ -r ] [ -u ] file ...\n", argv[0]);
	exit(1);
    }

    for (i = 1; i < argc; ++i) {
	if (strcmp(argv[i], "-u") == 0) {
	    offset = 4 - offset;
	    continue;
	}

	if (strcmp(argv[i], "-r") == 0) {
	    reorder = ~reorder;
	    continue;
	}

	infp = fopen(argv[i], "r");
	if (!infp) {
	    perror("fopen");
	    goto cleanup;
	}

	/* read and inspect file header (struct filehdr): */

	if (fread(&fhdr, 1, sizeof(fhdr), infp) != sizeof(fhdr)) {
	    fprintf(stderr, "%s: fread() from %s: file header short read\n",
		    argv[0], argv[i]);
	    goto cleanup;
	}

	if (fhdr.f_opthdr != sizeof(ohdr)) {
	    fprintf(stderr, "%s: %s has unexpected optional header size\n",
		    argv[0], argv[i]);
	    goto cleanup;
	}

	/* read and inspect optional header (struct aouthdr): */

	if (fread(&ohdr, 1, sizeof(ohdr), infp) != sizeof(ohdr)) {
	    fprintf(stderr,
		    "%s: fread() from %s: optional header short read\n",
		    argv[0], argv[i]);
	    goto cleanup;
	}

	/* read each section header looking for a text segment: */

	for (j = 0; j < fhdr.f_nscns; ++j) {
	    if (fread(&shdr, 1, sizeof(shdr), infp) != sizeof(shdr)) {
		fprintf(stderr,
			"%s: fread() from %s: section header short read\n",
			argv[0], argv[i]);
		goto cleanup;
	    }
	    next_section = ftell(infp);

	    if (shdr.s_flags & STYP_TEXT) {
		fseek(infp, shdr.s_scnptr, SEEK_SET);
		insts = malloc(shdr.s_size);
		if (fread(insts, 1, shdr.s_size, infp) != shdr.s_size) {
		    fprintf(stderr,
			    "%s: fread() from %s: section short read\n",
			    argv[0], argv[i]);
		    goto cleanup;
		}
		schedule_bb_ev4(insts, shdr.s_size/4, offset, reorder);
	    }
	    fseek(infp, next_section, SEEK_SET);
	}

      cleanup:
	if (infp) {
	    fclose(infp);
	}
	errno = 0;
    }
    return 0;
}
